"""Tools to use the Integral Equation Method (IEM) in the monostatic
(i.e. backscattering) case, when u = k*sin(th), case as described in
Fung et al. [1994].

Note: Argument "m" is subradar.Model class
"""

import numpy as np
from numpy import cos, exp, float64, inf, log10, sin
from scipy.special import factorial
from mpmath import nsum

from . import roughness
from .Classdef import Fresnel, Roughness, Signal


nan = float('nan')


class Common(Fresnel):
    """Common IEM relationships"""
    def __init__(self, **kwargs):
        Fresnel.__init__(self, **kwargs)

    def f(self):
        """f_pp terms"""
        return {'vv': 2*self.R['vv']/cos(self.th),
                'hh':-2*self.R['hh']/cos(self.th)}



class Large_S(Signal, Fresnel):
    """Large roughness single-scatter approximation [sec. 5.5]"""
    pass



class Large_M(Signal, Fresnel):
    """Large roughness multiple-scatter approximation [sec. 5.5]"""
    pass



class Small_S(Common, Signal, Fresnel, Roughness):
    """Small-to-moderate roughness single-scatter approximation [sec. 5.5]"""
    def __init__(self, **kwargs):
        Roughness.__init__(self, **kwargs)
        Fresnel.__init__(self, **kwargs)
        Signal.__init__(self, **kwargs)
        Common.__init__(self, **kwargs)


    def F_sum(self):
        """F_pp(-)+F_pp(+) terms"""

        vv = 2*sin(self.th)**2 * (1+self.R['vv'])**2 / cos(self.th) * \
              ( (1-1/self.ep) + \
                (self.mu*self.ep - sin(self.th)**2 - self.ep*cos(self.th)**2) /
                (self.ep*cos(self.th))**2
              )

        hh = -2*sin(self.th)**2 * (1+self.R['hh'])**2 / cos(self.th) * \
              ( (1-1/self.mu) + \
                (self.mu*self.ep - sin(self.th)**2 - self.mu*cos(self.th)**2) /
                (self.mu*cos(self.th))**2
              )
        return {'vv':vv, 'hh':hh}


    def I(self, n):
        """I_pp terms"""

        def _I(self, pp, n):
            return (2*self.wk_z)**n * self.f()[pp] * exp(-(self.sh*self.wk_z)**2) \
                    + self.wk_z**n * self.F_sum()[pp] /2

        return {'vv':_I(self, 'vv', n),
                'hh':_I(self, 'hh', n)}


    def nRCS(self, n='richardson+shanks', kind='isotropic gaussian', db=False):
        """Normalized Radar cross-section"""

        def _nRCS(self, pp, n, kind):
            n = float(n)
            return self.wk**2 * exp(-2*(self.sh*self.wk_z)**2) /2 * \
                   ( self.sh**(2*n) * np.abs( self.I(n)[pp] )**2 * \
                     roughness.spectrum(self.wk, self.sh, self.cl, self.th, \
                     n=n, kind=kind) / factorial(n) )

        if type(n) is not str:
            vv = nsum(lambda x: _nRCS(self, 'vv', x, kind), [1,n] )
            hh = nsum(lambda x: _nRCS(self, 'hh', x, kind), [1,n] )

        if type(n) is str:
            vv = nsum(lambda x: _nRCS(self, 'vv', x, kind), [1,inf], method=n)
            hh = nsum(lambda x: _nRCS(self, 'hh', x, kind), [1,inf], method=n)

        vv, hh = float64(vv), float64(hh)

        if db:
            vv, hh = 10*log10(vv), 10*log10(hh)

        return {'vv':vv, 'hh':hh}



class Small_M(Signal, Fresnel):
    """Small-to-moderate roughness multiple-scatter approximation [sec. 5.5]"""
    pass



"""Various Python classes"""

__author__ = 'Cyril Grima'

import numpy as np

from . import roughness, surface, utils

NAN = float('nan')


class Signal(object):
    """Signal relationships"""
    def __init__(self, wf=NAN, bw=NAN, th=0., bmw=NAN, h=NAN, **kwargs):
        self.wf = wf  # Signal central frequency [Hz]
        self.bw = bw  # Signal bandwidth [Hz]
        self.th = th  # Incident angle [rad]
        self.bmw = bmw  # Beamwidth [rad]
        self.h = h  # Altitude [m]

        # Defined from above variables
        self.wl = utils.wf2wl(wf)
        self.wk = utils.wf2wk(wf)
        self.wk_x = utils.wk2vec(self.wk, self.th)['x']
        self.wk_z = utils.wk2vec(self.wk, self.th)['z']
        self.footprint_rad = {'beam':utils.footprint_rad_beam(self.h, self.bmw),
                             'pulse':utils.footprint_rad_pulse(self.h, self.bw),
                             'fresnel':utils.footprint_rad_fresnel(self.h, self.wl)
                             }


class Fresnel(object):
    """Fresnel and Snell's relationships"""
    def __init__(self, ep1=1., ep2=1., mu1=1., mu2=1., th=0, **kwargs):
        self.ep1 = ep1  # Electric permittivity 1
        self.ep2 = ep2  # Electric permittivity 2
        self.ep = ep2/ep1  # Relative electric permittivity
        self.mu1 = mu1  # Magnetic permeability 1
        self.mu2 = mu2  # Magnetic permeability 2
        self.mu = mu2/mu1  # Relative magnetic permeability 2/1
        self.th = th  # Incident angle [rad]
        self.n1 = utils.epmu2n(ep1, mu1)
        self.n2 = utils.epmu2n(ep2, mu2)
        self.R = {'vv':utils.R_v(self.ep1, self.ep2, self.mu1, self.mu2, self.th),
                 'hh':utils.R_h(self.ep1, self.ep2, self.mu1, self.mu2, self.th),
                 'nn':utils.R(self.ep1, self.ep2, self.mu1, self.mu2, self.th)
                  }
        self.T = {'vv':utils.T_v(self.ep1, self.ep2, self.mu1, self.mu2, self.th),
                 'hh':utils.T_h(self.ep1, self.ep2, self.mu1, self.mu2, self.th)
                  }


class Roughness(object):
    """Roughness relationships"""
    def __init__(self, wf=NAN, sh=0, cl=np.inf, **kwargs):
        self.wf = wf  # Signal central frequency [Hz]
        self.sh = sh  # RMS  height [m]
        self.cl = cl  # Correlation Length [m]

        # Defined from above variables
        self.wl = utils.wf2wl(wf)
        self.wk = utils.wf2wk(wf)
        self.ks = self.wk*self.sh
        self.kl = self.wk*self.cl



"""Various tools fpr surface detection"""

import numpy as np
import pandas as pd


def detector(rdg, y0=[], winsize=100, method='grima2012', axis=0, **kwargs):
    """Surface detection with the choosen method

    Input
    -----
    rdg: 2d-array
        radargram.

    y0: array
        Initial estimation for the location of the surface.
        Optional.

    winsize: float
        Size of the window around y0 to look for the surface.
        Activated only if y0 > 0.

    method: string
        method to use for surface detection.

    axis: 0 or 1
        Long-time axis.

    Output
    ------
    y: float
        index of the location of the detected echo.

    """
    if axis == 1:
        rdg = np.rot90(rdg)

    xsize = rdg.shape[0]
    ysize = rdg.shape[1]
    y = np.zeros(xsize)

    # Detection
    for xi in np.arange(xsize):
        signal = rdg[xi,:]

        #index vector
        if len(y0) > 0:
            idx = np.arange(winsize)+y0[xi]-winsize/2.
        else:
            idx = np.arange(ysize)
        
        # Method selection
        if method == 'grima2012':
            y[xi], c = grima2012(signal, idx=idx, **kwargs)
        if method == 'mouginot2010':
            y[xi], c = mouginot2010(signal, idx=idx, **kwargs)

    return y


def mouginot2010(signal, idx=[], period=3, window=30, **kwargs):
    """Surface detection using [Mouginot et al. 2010]
    
    Parameters
    ----------
    signal: array
        signal vector

    idx: array
        the indices of the array where to search for the echo

    period: float
        window shift to compute the noise (=1 in the original paper)

    window: float
        size of the window where to compute the noise

    Output
    ------
    y: float
        index of the location of the detected echo

    c: array
        criteria computed with idx
    """

    # array of index where to search for the surface
    idx = np.array(idx)
    if idx.size == 0 :
        idx = np.arange(len(signal)).astype(int)
    else:
        idx = np.array(idx).astype(int) # make idx an integer array

    # Estimator calculation
    noise = pd.Series(signal[idx]).shift(periods=period).rolling(window).mean().values
    #noise = [np.nanmean(signal[i-30:i-3]) for i in idx]
    c = signal[idx]/noise

    # surface index
    try:
        y = idx[np.nanargmax(c)]
    except ValueError:
        y = np.nan
    return y, c


def grima2012(signal, idx=[], **kwargs):
    """Surface detection from [Grima et al. 2012]
    
    Parameters
    ----------
    signal: array
        signal vector

    idx: array
        the indices of the array where to search for the echo

    Output
    ------
    y: float
        index of the location of the detected echo

    c: array
        criteria computed with idx
    """

    # array of index where to search for the surface
    idx = np.array(idx)
    if idx.size == 0 :
        idx = np.arange(len(signal)).astype(int)
    else:
        idx = np.array(idx).astype(int) # make idx an integer array

    # Estimator calculation
    derivative = np.roll(np.gradient(signal[idx]), 2)
    c = signal[idx]*derivative

    # surface index
    try:
        y = idx[np.nanargmax(c)]
    except ValueError:
        y = np.nan
    return y, c



"""Various tools for surface desciption"""

import numpy as np

def spectrum(wk, sh, cl, th, n=1, kind='isotropic gaussian', **kwargs):
    """Surface roughness spectrum at at order n
    """

    if kind == 'isotropic gaussian':
        out = np.float64(cl)**2/n*np.exp(-(wk*cl*np.sin(th))**2/n) /2.
    else:
        raise ValueError("Unsupported value for kind: " + kind)

    return out



# This pip setup is made from instructions available at:
# https://towardsdatascience.com/how-to-build-your-first-python-package-6a00b02635c9

import pathlib
from setuptools import setup, find_packages
import subradar

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# Package information
PACKAGE_NAME = 'subradar'
AUTHOR = 'Cyril Grima'
AUTHOR_EMAIL = 'cyril.grima@gmail.com'
URL = 'https://github.com/cgrima/subradar'

LICENSE = 'MIT'
DESCRIPTION = 'Python Library for electromagnetic physics applied to subsurface radars'
LONG_DESCRIPTION = (HERE / "README.md").read_text()
LONG_DESC_TYPE = "text/markdown"

# Dependencies
INSTALL_REQUIRES = [
      'importlib>=1.0.4',
      'matplotlib>=3.3.3',
      'mpmath>=1.1.0',
      'numpy>=1.19.4',
      'pandas>=1.1.4',
      'scipy>=1.5.2',
]

# Bundle everything above
setup(name=PACKAGE_NAME,
      version=subradar.__version__,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      long_description_content_type=LONG_DESC_TYPE,
      author=AUTHOR,
      license=LICENSE,
      author_email=AUTHOR_EMAIL,
      url=URL,
      install_requires=INSTALL_REQUIRES,
      packages=find_packages()
      )
